require 'calibratewizard'
require 'cameradialog'
require 'controller'
require 'controlwidget'
require 'drb'
require 'hornetseye'
require 'preferencesdialog'
require 'qrc_visiongui'
require 'ui_clientwindow'

include Hornetseye

class ClientWindow < Qt::MainWindow

  VISIONDELAY = 500
  STATUSDELAY = 5000

  slots 'calibrate_drive()'
  slots 'open_camera()'
  slots 'next_frame()'
  slots 'vision(bool)'
  slots 'preferences()'

  def initialize( parent = nil )
    super parent

    DRb.start_service
    @controller = DRbObject.new nil, 'druby://:10000'
    @feedback = DRbObject.new nil, 'druby://:20000'
    calibrate_timing
    # @controller.home

    @ui = Ui::ClientWindow.new
    @ui.setupUi self

    @control_widget = ControlWidget.new @controller, @t_offset, @t_delta,
                                        nil, false
    @ui.scroll_area.widget = @control_widget

    connect @ui.action_quit, SIGNAL('triggered()'), self, SLOT('close()')
    connect @ui.action_opencamera, SIGNAL('triggered()'),
      self, SLOT('open_camera()')
    connect @ui.action_calibrate, SIGNAL('triggered()'),
      self, SLOT('calibrate_drive()')
    connect @ui.action_recognise, SIGNAL('toggled(bool)'),
      self, SLOT('vision(bool)')
    connect @ui.action_preferences, SIGNAL('triggered()'),
      self, SLOT('preferences()')
    connect @control_widget, SIGNAL('query_frame()'),
      self, SLOT('next_frame()')

    @video, @dt, @video_timer, @vision_timer, @vision_delay,
      @control_widget.delay, @frame =
      nil, nil, 0, 0, VISIONDELAY, VISIONDELAY, nil
  end

  def calibrate_timing
    count = 10
    samples = 500
    t = Time.new.to_f
    count.times { @controller.test samples }
    elapsed1 = ( Time.new.to_f - t ) / count
    t = Time.new.to_f
    count.times { @controller.test 2 * samples }
    elapsed2 = ( Time.new.to_f - t ) / count
    @t_offset = 2 * elapsed1 - elapsed2
    @t_delta = [ elapsed2 - elapsed1, 1.0e-8 ].max / samples
  end

  def calibrate_drive
    stop
    wizard = CalibrateWizard.new @video, @controller, @dt,
                                 @t_offset, @t_delta, self
    if wizard.exec == 0
      statusBar.showMessage 'Aborted calibration', STATUSDELAY
    else
      statusBar.showMessage 'Calibration completed', STATUSDELAY
      start wizard.xh, wizard.yh, wizard.zh, wizard.th
    end
  end

  def stop
    @ui.action_recognise.enabled = true
    @ui.action_recognise.checked = false
    @control_widget.deactivate
    if @video_timer != 0
      killTimer @video_timer
      @video_timer = 0
    end
  end

  def start( xh, yh, zh, th )
    @control_widget.activate @video.width, @video.height, xh, yh, zh, th
    @video_timer = startTimer 0 if @video_timer == 0
    @ui.action_recognise.enabled = true
  end

  def open_camera
    begin
      dialog = CameraDialog.new self
      if dialog.exec == 0
        statusBar.showMessage 'Aborted camera dialog', STATUSDELAY
      else
        statusBar.showMessage 'Accessing firewire digital camera', STATUSDELAY
        access_camera dialog.device, dialog.node, dialog.channel,
          dialog.format, dialog.mode, dialog.framerate, dialog.dt
      end
    rescue Exception => e
      statusBar.showMessage e.to_s, STATUSDELAY
      Qt::MessageBox::critical self, 'Error Accessing Camera', e.to_s
    end
  end

  def access_camera( device, node, channel, format, mode, framerate, dt )
    @video = DC1394Input.new device, node, channel, format, mode, -1,
                             framerate, DC1394Input::SPEED_400
    # @video = XineInput.new 'test.avi', true
    @dt, @control_widget.dt = dt, dt
    @control_widget.resize @video.width, @video.height
    @ui.action_opencamera.enabled = false
    @ui.action_calibrate.enabled = true
  end

  def timerEvent( e )
    case e.timerId
    when @video_timer
      next_frame
    when @vision_timer
      result = @feedback.stream @frame.width, @frame.height,
                                @frame.to_ubyte.export
      @control_widget.feedback = result
    end
  end

  def next_frame
    @frame = @video.read
    @control_widget.write @frame
  end

  def vision( on )
    @ui.action_preferences.enabled = !on
    if on
      @vision_timer = startTimer @vision_delay if @vision_timer == 0
    elsif @vision_timer != 0
      killTimer @vision_timer
      @vision_timer = 0
      @control_widget.feedback = nil
    end
  end

  def preferences
    dialog = PreferencesDialog.new self
    dialog.vision_delay = @vision_delay * 0.001
    dialog.path_speed = @control_widget.speed
    if dialog.exec == Qt::Dialog::Accepted
      @vision_delay = ( dialog.vision_delay * 1000 ).to_i
      @control_widget.speed = dialog.path_speed
      @control_widget.delay = @vision_delay
      statusBar.showMessage 'Preferences changed', STATUSDELAY
    else
      statusBar.showMessage 'Aborted preferences dialog', STATUSDELAY
    end
  end

end
