require 'controller'
require 'hornetseye'

include Hornetseye

class ControlWidget < XvWidget

  PATHSPEED   = 100

  signals 'query_frame()'

  attr_accessor :dt, :speed, :delay

  def initialize( controller, t_offset, t_delta,
                  parent = nil, grab_port_now = true )
    super parent, grab_port_now
    @vx, @vy = nil, nil
    @target, @nominal, @old_nominal, @feedback, @deviation =
      nil, nil, nil, nil, nil
    @controller, @dt, @delay = controller, nil, nil
    @speed, @t_offset, @t_delta = PATHSPEED, t_offset, t_delta
    @active = false
    @control_timer = 0
  end

  def mousePressEvent( e )
    if @active
      @nominal, @old_nominal = @feedback, @feedback
      @target = [ e.x, e.y ]
      update
    end
  end

  def mouseMoveEvent( e )
    if @nominal.nil? and @active
      if @target
        drive e.x - @target[0], e.y - @target[1]
      end
      @target = [ e.x, e.y ]
      emit query_frame
    end
  end

  def mouseReleaseEvent( e )
    if @active
      @nominal, @old_nominal = nil, nil
      if @control_timer != 0
        killTimer @control_timer
        @control_timer = 0
      end
      update
    end
  end

  def activate( w, h, xh, yh, zh, th )
    grabPort
    sx, sy = 2.0 / w, 2.0 / h
    xdrift = ( xh[ 3 ] - xh[ 0 ] ) / ( th[ 3 ] - th[ 0 ] )
    ydrift = ( yh[ 3 ] - yh[ 0 ] ) / ( th[ 3 ] - th[ 0 ] )
    zdrift = ( zh[ 3 ] - zh[ 0 ] ) / ( th[ 3 ] - th[ 0 ] )
    @vx = [ ( xh[ 1 ] - xdrift * ( th[ 1 ] - th[ 0 ] ) - xh[ 0 ] ) * sx,
            ( yh[ 1 ] - ydrift * ( th[ 1 ] - th[ 0 ] ) - yh[ 0 ] ) * sx,
            ( zh[ 1 ] - zdrift * ( th[ 1 ] - th[ 0 ] ) - zh[ 0 ] ) * sx ]
    @vy = [ ( xh[ 2 ] - xdrift * ( th[ 2 ] - th[ 0 ] ) - xh[ 1 ] ) * sy,
            ( yh[ 2 ] - ydrift * ( th[ 2 ] - th[ 0 ] ) - yh[ 1 ] ) * sy,
            ( zh[ 2 ] - zdrift * ( th[ 2 ] - th[ 0 ] ) - zh[ 1 ] ) * sy ]
    @active = true
  end

  def deactivate
    @active = false
    releasePort
  end

  def drive( dx, dy )
    pos = @controller.pos
    rx = pos[ 0 ] + @vx[ 0 ] * dx + @vy[ 0 ] * dy
    ry = pos[ 1 ] + @vx[ 1 ] * dx + @vy[ 1 ] * dy
    rz = pos[ 2 ] + @vx[ 2 ] * dx + @vy[ 2 ] * dy
    @controller.drive rx, ry, rz, ( ( @dt - @t_offset ) / @t_delta ).to_i
  end

  def feedback=( value )
    @feedback = value
    if @feedback
      if @nominal and @control_timer == 0
        @control_timer = startTimer( ( @dt * 1000 ).to_i )
      end
      if @old_nominal
        @deviation = [ @old_nominal[ 0 ] - @feedback[ 0 ],
                       @old_nominal[ 1 ] - @feedback[ 1 ] ]
      else
        @deviation = nil
      end
    else
      if @control_timer != 0
        killTimer @control_timer
        @control_timer = 0
      end
      @deviation = nil
    end
    @old_nominal = @nominal
    update
  end

  def paintEvent( e )
    super e
    p = Qt::Painter.new self
    p.setRenderHint Qt::Painter::Antialiasing, false
    p.pen = Qt::Pen.new Qt::Color.new( 255, 0, 0 )
    if @feedback
      cx, cy = @feedback[ 0 ], @feedback[ 1 ]
      p.drawEllipse cx - 5, cy - 5, 10, 10
      p.drawLine cx - 5, cy, cx + 5, cy
      p.drawLine cx, cy - 5, cx, cy + 5
    end
    if @target
      if @nominal
        cx, cy = @nominal[ 0 ], @nominal[ 1 ]
        p.drawLine cx - 5, cy, cx + 5, cy
        p.drawLine cx, cy - 5, cx, cy + 5
        p.drawLine cx, cy, @target[ 0 ], @target[ 1 ]
      end
      cx, cy = @target[ 0 ], @target[ 1 ]
      p.drawEllipse cx - 5, cy - 5, 10, 10
    end
    p.end
  end

  def timerEvent( e )
    if e.timerId == @control_timer
      diff = [ @target[ 0 ] - @nominal[ 0 ],
               @target[ 1 ] - @nominal[ 1 ] ]
      dist = Math.sqrt( diff[0] ** 2 + diff[1] ** 2 )
      if @deviation
        factor = 1000 * @dt / @delay
        correction = [ @deviation[ 0 ] * factor,
                       @deviation[ 1 ] * factor ]
      else
        correction = [ 0, 0 ]
      end
      if dist < @speed * @dt
        f = 1
      else
        f = @speed * @dt / dist
      end
      dx = diff[ 0 ] * f
      dy = diff[ 1 ] * f
      @nominal = [ @nominal[ 0 ] + dx, @nominal[ 1 ] + dy ]
      drive dx + correction[ 0 ], dy + correction[ 1 ]
    end
  end

end
