/* Copyright (C) 2006 Jan Wedekind.
   This file is part of the computer vision library Hornetseye.

   HornetsEye is free software; you can redistribute it and/or modify it under
   the terms of the GNU GENERAL PUBLIC LICENSE as published by the Free
   Software Foundation; either version 2 of the License, or (at your option)
   any later version.

   HornetsEye is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTIBILITY or FITNESS
   FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
   details.

   You should have received a copy of the GNU General Public License along
   with HornetsEye; if not, contact one of the authors of this software. */
#ifndef HORNETSEYE_BASE_ERROR_HH
#define HORNETSEYE_BASE_ERROR_HH

#include <exception>
#include <sstream>
#include <string>

namespace hornetseye {

  /** Exception class deriving from std::exception.
      This class provides a syntax similar to output-streams for convenience.
      For compability with other libraries it is inheriting the class
      \c std::exception.
    
      Here is an example how to use an instance of this class in C++:
      \include exceptiontest/exceptiontest.cc
      
      Exception-handling also can be done using the macro \c ERRORMACRO:
      \include exceptiontest/macrotest.cc
      
      Mind that the macro uses a variable with the name \c _e. Make sure, that
      you don't use this variable-name in any of the macro-arguments!

      Ruby already comes with exception classes:
      \include exceptiontest/exceptiontest.rb
      
      @author Jan Wedekind (jan at wedesoft.de)
      @date Mon Aug 23 14:37:05 UTC 2004 */
  class Error: public std::exception
  {
  public:
    /// Constructor.
    Error(void) {}
    /// Copy constructor.
    Error( Error &e ): std::exception( e )
      { m_message << e.m_message.str(); }
    /// Destructor.
    virtual ~Error(void) throw() {}
    ///
    template< typename T >
    std::ostream &operator<<( const T &t )
      { m_message << t; return m_message; }
    /** Interface for manipulators.
        Manipulators such as \c std::endl and \c std::hex use these
        functions in constructs like "Error e; e << std::endl".
        For more information, see the iomanip header. */
    std::ostream &operator<<( std::ostream& (*__pf)( std::ostream&) )
      { (*__pf)( m_message ); return m_message; }
    /// Returns error message (not thread safe).
    virtual const char* what(void) const throw() {
      temp = m_message.str();
      return temp.c_str();
    }
  protected:
    /// Memory-stream containing the error message.
    std::ostringstream m_message;
    /** Temporary to do null-termination.
        The method \c what() requires a null-terminated string. */
    static std::string temp;
  };
  
};

#define ERRORMACRO( condition, class, params, message ) \
  if ( !( condition ) ) {                               \
    class _e params;                                    \
    _e << message;                                      \
    throw _e;                                           \
  };

#endif
