require 'connected_components'
require 'visionbase'
require 'ui_plugin_cc'

class ComponentsWidget < ImageDisplay

  signals 'roi(int,int,int,int)'
  signals 'noroi()'

  def initialize( parent = nil )
    super parent
    @selection, @selecting, @image = false, false, nil
    @x0, @y0, @w, @h = 0, 0, 0, 0
    @cross, @x, @y = false, 0, 0
  end

  def write( image )
    @image = image
    super image
  end

  def clear_cross
    @cross = false
    repaint @x - 5, @y - 5, 11, 11
    @x, @y = 0, 0
  end

  def set_cross( x, y )
    clear_cross
    @cross, @x, @y = true, x, y
    repaint @x - 5, @y - 5, 11, 11
  end

  def paintEvent( e )
    super e
    p = Qt::Painter.new self
    p.setRenderHint Qt::Painter::Antialiasing, false
    if @cross
      pen = Qt::Pen.new Qt::Color.new( 255, 0, 0 )
      p.pen = pen
      p.drawLine @x - 5, @y, @x + 5, @y
      p.drawLine @x, @y - 5, @x, @y + 5
    end
    if @selection or @selecting
      pen = Qt::Pen.new Qt::Color.new( 255, 0, 0 )
      pen.style = Qt::DotLine if @selecting
      p.pen = pen
      p.drawRect @x0, @y0, @w - 1, @h - 1
    end
    p.end
  end

  def mousePressEvent( e )
    if @image != nil
      case e.button
      when Qt::LeftButton
        clear_roi
        @x0, @y0, @w, @h = e.x, e.y, 1, 1
        @selection, @selecting = false, true
        repaint @x0, @y0, @w, @h
      when Qt::RightButton
        unless @selecting
          clear_roi
          emit noroi
        end
      end
    end
  end

  def mouseMoveEvent( e )
    if @selecting
      x = [ @x0, @x0 + @w - 1 ]
      y = [ @y0, @y0 + @h - 1 ]
      @w = [ [ @image.shape[0] - @x0, e.x + 1 - @x0 ].min, 1 ].max
      @h = [ [ @image.shape[1] - @y0, e.y + 1 - @y0 ].min, 1 ].max
      x.push @x0 + @w - 1
      y.push @y0 + @h - 1
      repaint x.min, y.min, x.max + 1 - x.min, y.max + 1 - y.min
    end
  end

  def mouseReleaseEvent( e )
    if @selecting
      case e.button
      when Qt::LeftButton
        @selection, @selecting = true, false
        repaint @x0, @y0, @w, @h
        emit roi( @x0, @y0, @w, @h )
      end
    end
  end

  def clear_roi
    if @selection or @selecting
      x = [ @x0, @x0 + @w - 1 ]
      y = [ @y0, @y0 + @h - 1 ]
      @selection, @selecting = false, false
      repaint @x0, @y0, @w, @h
    end
  end

  def set_roi( x, y, w, h )
    clear_roi
    @x0, @y0, @w, @h = x, y, w, h
    @selection, @selecting = true, false
  end

end

class FormCc < AlgorithmForm

  slots 'lower(bool)'
  slots 'upper(bool)'
  slots 'lower_threshold(int)'
  slots 'upper_threshold(int)'
  slots 'erosion(int)'
  slots 'area(int)'
  slots 'angle(int)'
  slots 'noroi()'
  slots 'roi(int,int,int,int)'

  def initialize( parent = nil )
    super parent
    @ui = Ui::FormCc.new
    @ui.setupUi self
    @components_widget = ComponentsWidget.new
    @ui.scroll_area.widget = @components_widget
    connect @ui.lower_check, SIGNAL('toggled(bool)'),
      self, SLOT('lower(bool)')
    connect @ui.upper_check, SIGNAL('toggled(bool)'),
      self, SLOT('upper(bool)')
    connect @ui.lower_spin, SIGNAL('valueChanged(int)'),
      self, SLOT('lower_threshold(int)')
    connect @ui.upper_spin, SIGNAL('valueChanged(int)'),
      self, SLOT('upper_threshold(int)')
    connect @ui.erode_spin, SIGNAL('valueChanged(int)'),
      self, SLOT('erosion(int)')
    connect @ui.area_spin, SIGNAL('valueChanged(int)'),
      self, SLOT('area(int)')
    connect @ui.angle_dial, SIGNAL('valueChanged(int)'),
      self, SLOT('angle(int)')
    connect @components_widget, SIGNAL('noroi()'),
      self, SLOT('noroi()')
    connect @components_widget, SIGNAL('roi(int,int,int,int)'),
      self, SLOT('roi(int,int,int,int)')
    @image, @timer = nil, nil
  end

  def image=( image )
    @components_widget.write image
    @image = image
  end

  def algorithm=( algorithm )
    super algorithm
    image, @image = @image, nil
    lower = algorithm.lower_threshold
    @ui.lower_spin.value = lower != nil ? lower : 0
    @ui.lower_check.checked = lower != nil
    upper = algorithm.upper_threshold
    @ui.upper_spin.value = upper != nil ? upper : 255
    @ui.upper_check.checked = upper != nil
    @ui.erode_spin.value = algorithm.erode
    enabled = lower != nil || upper != nil
    @ui.erode_spin.enabled = enabled
    @ui.area_spin.enabled = enabled
    @ui.angle_dial.enabled = enabled
    @ui.angle_dial.value = @algorithm.angle
    @ui.area_spin.value = algorithm.area
    if @algorithm.visible
      pose = algorithm.model.pose * algorithm.delta.inv
      @components_widget.set_cross pose[ 0, 2 ], pose[ 1, 2 ]
    else
      @components_widget.clear_cross
    end
    if @algorithm.roi
      @components_widget.set_roi @algorithm.x, @algorithm.y,
        @algorithm.w, @algorithm.h
    else
      @components_widget.clear_roi
    end
    @image = image
  end

  def lower( on )
    @ui.lower_spin.enabled = on
    enable = @ui.lower_spin.enabled || @ui.upper_spin.enabled
    @ui.erode_spin.enabled = enable
    @ui.area_spin.enabled = enable
    @ui.angle_dial.enabled = enabled
    @algorithm.lower_threshold = on ? @ui.lower_spin.value : nil
    update_cc
  end

  def upper( on )
    @ui.upper_spin.enabled = on
    enable = @ui.lower_spin.enabled || @ui.upper_spin.enabled
    @ui.erode_spin.enabled = enable
    @ui.area_spin.enabled = enable
    @ui.angle_dial.enabled = enabled
    @algorithm.upper_threshold = on ? @ui.upper_spin.value : nil
    update_cc
  end

  def lower_threshold( value )
    if @ui.lower_spin.enabled
      @algorithm.lower_threshold = value
      update_cc
    end
  end

  def upper_threshold( value )
    if @ui.upper_spin.enabled
      @algorithm.upper_threshold = value
      update_cc
    end
  end

  def erosion( value )
    @algorithm.erode = value
    update_cc
  end

  def area( value )
    @algorithm.area = value
    update_cc
  end

  def angle( value )
    @algorithm.angle = value
    update_cc
  end

  def noroi
    @algorithm.clear_roi
    update_cc
  end

  def roi( x, y, w, h )
    @algorithm.set_roi x, y, w, h
    update_cc
  end

  def update_cc
    if @image != nil
      if @timer != nil
        killTimer @timer
        @timer = nil
      end
      labels = @algorithm.update @image
      @components_widget.write labels.normalise
      if @algorithm.visible
        pose = @algorithm.model.pose * @algorithm.delta.inv
        @components_widget.set_cross pose[ 0, 2 ], pose[ 1, 2 ]
      else
        @components_widget.clear_cross
      end
      @timer = startTimer 1000
    end
  end

  def timerEvent( e )
    if e.timerId == @timer
      @components_widget.write @image unless @image == nil
      killTimer @timer
      @timer = nil
    end
  end

end

class AlgorithmCc < Recognition

  class << self

    def description
      'Connected Components Analysis'
    end

    def config_widget( parent = nil )
      FormCc.new parent
    end

  end

  attr_accessor :lower_threshold, :upper_threshold, :erode, :area, :angle,
    :visible, :x, :y, :w, :h, :roi

  def initialize( model = nil )
    super model
    @lower_threshold, @upper_threshold, @erode, @area, @angle, @visible =
      nil, nil, 1, 0, 0, false
    @roi, @x, @y, @w, @h = false, 0, 0, 0, 0
  end

  Struct.new 'AlgorithmCc', :base, :lower_threshold, :upper_threshold, :erode,
    :area, :angle, :visible, :roi, :x, :y, :w, :h

  def marshal_load( str )
    super str.base
    @lower_threshold, @upper_threshold, @erode, @area, @angle, @visible =
      str.lower_threshold, str.upper_threshold, str.erode, str.area,
      str.angle, str.visible
    @roi, @x, @y, @w, @h = str.roi, str.x, str.y, str.w, str.h
    self
  end

  def marshal_dump
    Struct::AlgorithmCc.new super, @lower_threshold, @upper_threshold, @erode,
      @area, @angle, @visible, @roi, @x, @y, @w, @h
  end

  def clear_roi
    @roi = false
  end

  def set_roi( x, y, w, h )
    @x, @y, @w, @h, @roi = x, y, w, h, true
  end

  def overlay( widget )
    if @visible
      p = Qt::Painter.new widget
      p.setRenderHint Qt::Painter::Antialiasing, false
      pen = Qt::Pen.new Qt::Color.new( 255, 0, 0 )
      pen.style = Qt::DotLine unless success?
      p.pen = pen
      p.setWorldTransform( ( @model.pose * @delta.inv ).to_qtransform )
      p.drawLine -5, 0, 5, 0
      p.drawLine 0, -5, 0, 5
      p.end
    end
  end

  def success?
    @visible
  end

  def update( image )
    lower = @lower_threshold ? @lower_threshold : 0
    upper = @upper_threshold ? @upper_threshold : 255
    if @roi
      region = image[ @x ... @x + @w, @y ... @y + @h ]
    else
      region = image
    end
    selection = region.between? lower, upper
    if @erode > 1
      s = 2 * ( @erode - 1 ) + 1
      xerode = MultiArray.to_multiarray [ [1] * s ]
      yerode = MultiArray.to_multiarray [ [1] ] * s
      selection = selection.to_ubyte.
        correlate( xerode ).correlate( yerode ).eq s ** 2
    end
    # Compute labels
    labels = selection.to_ubyte.labeling_v1
    # Filter small areas
    min_area = region.size * @area / 1000
    mask = labels.hist( labels.max + 1 ) >= min_area
    reindex = mask.to_ubyte.integral * mask.to_ubyte
    labels = labels.map reindex
    # Sort areas
    n = reindex.max + 1
    if n > 1
      areas = labels.hist n
      idx = MultiArray.int( *labels.shape ).indgen!
      x, y = idx % labels.shape[0], idx / labels.shape[0]
      a = @angle * Math::PI / 180
      weights = y * -Math.cos( a ) + x * -Math.sin( a )
      order = labels.hist_weighted( n, weights )[ 1 ... n ] / areas[ 1 ... n ]
      sorted = Sequence.to_sequence( [0] + ( 1 ... n ).zip( order.to_a ).
                                     sort_by { |i,o| o }.collect { |i,o| i } )
      labels = labels.map sorted.inverse
      selected = labels.eq n - 1
      oriented = weights.mask selected
      primary = idx.mask( selected ).mask( oriented <= oriented.min )
      if primary.size >= 1
        x0, y0 = primary[0] % labels.shape[0], primary[0] / labels.shape[0]
        if @roi
          x0 += @x
          y0 += @y
        end
        @model.pose = Pose.new( x0, y0 ) * @delta
        @visible = true
      else
        @visible = false
      end
    else
      @visible = false
    end
    if @roi
      retval = MultiArray.new( labels.typecode, *image.shape ).fill!
      retval[ @x ... @x + @w, @y ... @y + @h ] = labels
      retval
    else
      labels
    end
  end

end
