require 'visionbase'
require 'ui_plugin_cog'

class PickDisplay < ImageDisplay

  signals 'picked(int)'

  def initialize( parent = nil )
    super parent
    @image = nil
    @x, @y, @n, @n_min = 0, 0, 0, 1
  end

  def write( image )
    super image
    @image = image
  end

  def set_cross( x, y, n, n_min )
    clear_cross
    @x, @y, @n, @n_min = x, y, n, n_min
    repaint @x - 5, @y - 5, 11, 11
  end

  def clear_cross
    @n, @n_min = 0, 1
    repaint @x - 5, @y - 5, 11, 11
    @x, @y = 0, 0
  end

  def paintEvent( e )
    super e
    if @n > 0
      p = Qt::Painter.new self
      p.setRenderHint Qt::Painter::Antialiasing, false
      pen = Qt::Pen.new Qt::Color.new( 255, 0, 0 )
      pen.style = Qt::DotLine if @n < @n_min
      p.pen = pen
      p.translate @x, @y
      p.drawLine -5, 0, 5, 0
      p.drawLine 0, -5, 0, 5
      p.end
    end
  end

  def mousePressEvent( e )
    if @image != nil
      if e.button == 1
        if e.x >= 0 and e.y >= 0 and
            e.x < @image.shape[0] and e.y < @image.shape[1]
          emit picked( @image[ e.x, e.y ] )
        end
      end
    end
  end

end

class FormCog < AlgorithmForm

  slots 'brightness(int)'
  slots 'threshold(int)'
  slots 'area(double)'

  def initialize( parent = nil )
    super parent
    @ui = Ui::FormCog.new
    @ui.setupUi self
    @pick_widget = PickDisplay.new
    @ui.scroll_area.widget = @pick_widget
    connect @ui.brightness_spin, SIGNAL('valueChanged(int)'),
      self, SLOT('brightness(int)')
    connect @ui.threshold_spin, SIGNAL('valueChanged(int)'),
      self, SLOT('threshold(int)')
    connect @ui.area_spin, SIGNAL('valueChanged(double)'),
      self, SLOT('area(double)')
    connect @pick_widget, SIGNAL('picked(int)'),
      @ui.brightness_spin, SLOT('setValue(int)')
    @image = nil
  end

  def image=( image )
    @pick_widget.write image
    @image = image
  end

  def algorithm=( algorithm )
    super algorithm
    image, @image = @image, nil
    @ui.brightness_spin.value = algorithm.brightness
    @ui.threshold_spin.value = algorithm.threshold
    pose = @algorithm.model.pose * @algorithm.delta.inv
    @pick_widget.set_cross pose[ 0, 2 ], pose[ 1, 2 ],
                           @algorithm.n, @algorithm.n_min
    @image = image
  end

  def brightness( value )
    @algorithm.brightness = value
    update_cog
  end

  def threshold( value )
    @algorithm.threshold = value
    update_cog
  end

  def area( value )
    @algorithm.area = value
    update_cog
  end

  def update_cog
    if @image != nil
      @algorithm.update @image
      @algorithm.n_min = [ @algorithm.n * @algorithm.area, 1 ].max
      pose = @algorithm.model.pose * @algorithm.delta.inv
      @pick_widget.set_cross pose[ 0, 2 ], pose[ 1, 2 ],
                             @algorithm.n, @algorithm.n_min
    end
  end

end

class AlgorithmCog < Recognition

  class << self

    def description
      'Center of Gravity'
    end

    def config_widget( parent = nil )
      FormCog.new parent
    end

  end

  attr_accessor :brightness, :threshold, :area, :n_min, :n

  def initialize( model = nil )
    super model
    @brightness, @threshold, @area = 128, 10, 0.9
    @n, @n_min = 0, 1
  end

  Struct.new 'AlgorithmCog', :base, :brightness, :threshold, :area, :n_min, :n

  def marshal_load( str )
    super str.base
    @brightness, @threshold, @area, @n_min, @n =
      str.brightness, str.threshold, str.area, str.n_min, str.n
    self
  end

  def marshal_dump
    Struct::AlgorithmCog.new super, @brightness, @threshold, @area,
      @n_min, @n
  end

  def overlay( widget )
    if @n >= 1
      p = Qt::Painter.new widget
      p.setRenderHint Qt::Painter::Antialiasing, false
      pen = Qt::Pen.new Qt::Color.new( 255, 0, 0 )
      pen.style = Qt::DotLine unless success?
      p.pen = pen
      p.setWorldTransform( ( @model.pose * @delta.inv ).to_qtransform )
      p.drawLine -5, 0, 5, 0
      p.drawLine 0, -5, 0, 5
      p.end
    end
  end

  def success?
    @n >= @n_min
  end

  def update( image )
    mask = ( image.to_sint - @brightness ).abs <= @threshold
    idx = MultiArray.int( *image.shape ).indgen!.mask( mask )
    @n = idx.size
    if @n > 0
      x0 = ( idx % image.shape[0] ).sum / @n
      y0 = ( idx / image.shape[0] ).sum / @n
      @model.pose = Pose.new( x0, y0 ) * @delta
    end
  end

end
