require 'Qt4'
require 'matrix'
require 'visionbase'
require 'ui_plugin_lk'

class AreaWidget < ImageDisplay

  signals 'capture(double,double,double)'

  def initialize( parent = nil )
    super parent
    @selection, @selecting, @image = false, false, nil
    @x0, @y0, @a, @w, @h, @fringe = 0, 0, 0.0, 0, 0, 5
  end

  def write( image )
    @image = image
    super image
  end

  def paintEvent( e )
    super e
    if @selection or @selecting
      p = Qt::Painter.new self
      p.setRenderHint Qt::Painter::Antialiasing, true
      p.setWorldTransform Pose.new( @x0, @y0, @a ).to_qtransform
      pen = Qt::Pen.new Qt::Color.new( 255, 0, 0 )
      pen.style = Qt::DotLine if @selecting
      p.pen = pen
      p.drawRect 0, 0, @w, @h
      if @selection
        pen = Qt::Pen.new Qt::Color.new( 255, 0, 0 )
        pen.style = Qt::DotLine
        p.pen = pen
        p.drawRect -@fringe, -@fringe, @w + 2 * @fringe, @h + 2 * @fringe
      end
      p.end
    end
  end

  def mousePressEvent( e )
    if @image != nil
      if @image.shape[0] > 2 * @fringe and @image.shape[1] > 2 * @fringe
        case e.button
        when Qt::LeftButton
          clear
          @x0, @y0, @a, @w, @h = e.x, e.y, 0.0, 1, 1
          @selection, @selecting = false, true
          repaint
        end
      end
    end
  end

  def mouseMoveEvent( e )
    if @selecting
      @w = [ [ @image.shape[0] - @x0, e.x + 1 - @x0 ].min, 1 ].max
      @h = [ [ @image.shape[1] - @y0, e.y + 1 - @y0 ].min, 1 ].max
      repaint
    end
  end

  def mouseReleaseEvent( e )
    if @selecting
      case e.button
      when Qt::LeftButton
        @selection, @selecting = true, false
        repaint
        emit capture( @x0, @y0, @a )
      end
    end
  end

  def wheelEvent( e )
    if @selection
      delta = e.delta / 8.0 * 2 * Math::PI / 360.0
      self.a = @a + delta
    end
  end

  def a=( value )
    @x0 += Math.cos( @a ) * @w * 0.5 - Math.sin( @a ) * @h * 0.5
    @y0 += Math.sin( @a ) * @w * 0.5 + Math.cos( @a ) * @h * 0.5
    @a = value % ( 2 * Math::PI )
    @x0 -= Math.cos( @a ) * @w * 0.5 - Math.sin( @a ) * @h * 0.5
    @y0 -= Math.sin( @a ) * @w * 0.5 + Math.cos( @a ) * @h * 0.5
    repaint
    emit capture( @x0, @y0, @a )
  end

  def selection( x0, y0, a, w, h )
    @x0, @y0, @a, @w, @h =  x0, y0, a, w, h
    @selection = true
    repaint
  end

  def fringe=( value )
    fringe_max = [ @fringe, value ].max
    @fringe = value
    if @selection
      repaint
    end
  end

  def clear
    if @selection or @selecting
      @selection, @selecting = false, false
      repaint
    end
  end

  def template
    if @selection
      w, h = @w + 2 * @fringe, @h + 2 * @fringe
      idx = MultiArray.int( w, h ).indgen!
      x = ( idx % w ) - @fringe
      y = ( idx / w ) - @fringe
      warp = MultiArray.int( w, h, 2 )
      warp[ 0...w, 0...h, 0 ] = Math.cos( @a ) * x - Math.sin( @a ) * y + @x0
      warp[ 0...w, 0...h, 1 ] = Math.sin( @a ) * x + Math.cos( @a ) * y + @y0
      @image.warp_clipped warp
    else
      nil
    end
  end

end

class FormLk < AlgorithmForm

  slots 'dof(int)'
  slots 'sigma(double)'
  slots 'iterations(int)'
  slots 'threshold(double)'
  slots 'fringe(int)'
  slots 'capture(double,double,double)'

  def initialize( parent = nil )
    super parent
    @ui = Ui::FormLk.new
    @ui.setupUi self
    @area_widget = AreaWidget.new
    @ui.scroll_area.widget = @area_widget
    connect @ui.dof_combo, SIGNAL('currentIndexChanged(int)'),
      self, SLOT('dof(int)')
    connect @ui.sigma_spin, SIGNAL('valueChanged(double)'),
      self,  SLOT('sigma(double)')
    connect @ui.iterations_spin, SIGNAL('valueChanged(int)'),
      self, SLOT('iterations(int)')
    connect @ui.threshold_spin, SIGNAL('valueChanged(double)'),
      self, SLOT('threshold(double)')
    connect @ui.fringe_spin, SIGNAL('valueChanged(int)'),
      self,  SLOT('fringe(int)')
    connect @area_widget, SIGNAL('capture(double,double,double)'),
      self, SLOT('capture(double,double,double)')
  end

  def image=( image )
    @area_widget.write image
  end

  def algorithm=( algorithm )
    super algorithm
    @ui.dof_combo.currentIndex = [ :dof2, :dof3 ].index @algorithm.mode
    @ui.sigma_spin.value = algorithm.sigma
    @ui.iterations_spin.value = algorithm.iterations
    @ui.threshold_spin.value = algorithm.threshold
    @ui.fringe_spin.value = algorithm.fringe
    if algorithm.template != nil
      pose = @algorithm.model.pose * @algorithm.delta.inv
      @area_widget.selection pose[ 0, 2 ], pose[ 1, 2 ],
        Math.atan2( pose[ 1, 0 ], pose[ 0, 0 ] ),
        algorithm.template.shape[0] - 2 * algorithm.fringe,
        algorithm.template.shape[1] - 2 * algorithm.fringe
    else
      @area_widget.clear
    end
  end

  def dof( value )
    @algorithm.mode = [ :dof2, :dof3 ][ value ]
    @area_widget.a = 0.0 unless value == 1
  end

  def sigma( value )
    @algorithm.sigma = value
  end

  def iterations( value )
    @algorithm.iterations = value
  end

  def threshold( value )
    @algorithm.threshold = value
  end

  def fringe( value )
    @algorithm.fringe = value
    @area_widget.fringe = value
    @algorithm.template = @area_widget.template
    @algorithm.deviation = 0.0
  end

  def capture( x, y, a )
    @ui.dof_combo.currentIndex = 1 unless a == 0.0
    @algorithm.template = @area_widget.template
    @algorithm.delta = Pose.new( x, y, a ).inv * @algorithm.model.pose
    @algorithm.deviation = 0.0
  end

end

class AlgorithmLk < Tracking

  class << self

    def description
      'Lucas-Kanade tracker'
    end

    def config_widget( parent = nil )
      FormLk.new parent
    end

  end

  attr_accessor :sigma, :iterations, :threshold, :template, :fringe,
    :mode, :deviation

  def initialize( model = nil )
    super model
    @sigma, @iterations, @threshold, @fringe, @template, @mode =
      2.5, 5, 32, 5, nil, :dof2
    @deviation = 0.0
  end

  Struct.new 'AlgorithmLk', :base, :sigma, :iterations, :threshold, :typecode,
    :export, :shape, :fringe, :mode, :deviation

  def marshal_load( str )
    super str.base
    @sigma, @iterations, @threshold, @fringe, @mode, @deviation =
      str.sigma, str.iterations, str.threshold, str.fringe, str.mode,
      str.deviation
    @template = MultiArray.import str.typecode, str.export, *str.shape
    self
  end

  def marshal_dump
    Struct::AlgorithmLk.new super, @sigma, @iterations, @threshold,
      @template.typecode, @template.export, @template.shape, @fringe, @mode,
      @deviation
  end

  def overlay( widget )
    if @template != nil
      p = Qt::Painter.new widget
      p.setRenderHint Qt::Painter::Antialiasing, false
      pen = Qt::Pen.new Qt::Color.new( 255, 0, 0 )
      pen.style = Qt::DashDotLine
      p.pen = pen
      p.setWorldTransform( ( @model.pose * @delta.inv ).to_qtransform )
      w, h = @template.shape[0] - 2 * @fringe, @template.shape[1] - 2 * @fringe
      p.drawRect 0, 0, w - 1, h - 1
      p.end
    end
  end

  def success?
    @deviation <= @threshold
  end

  def derivative( x, y )
    case @mode
    when :dof2
      Matrix[ [ 1, 0 ], [ 0, 1 ] ]
    when :dof3
      Matrix[ [ 1, 0 ], [ 0, 1 ], [ -y, x ] ]
    else
      raise "Unknown mode #{@mode}"
    end
  end

  def compose( pose, parameters )
    case @mode
    when :dof2
      dx, dy = *parameters
      Pose.new( pose[ 0, 2 ] + dx, pose[ 1, 2 ] + dy )
    when :dof3
      dx, dy, da = *parameters
      sw, cw = pose[ 1, 0 ], pose[ 0, 0 ]
      angle = Math.atan2( sw, cw )
      Pose.new( pose[ 0, 2 ] + dx * cw - dy * sw,
                pose[ 1, 2 ] + dx * sw + dy * cw,
                angle + da )
    else
      raise "Unknown mode #{@mode}"
    end
  end

  def update( image )
    if @template != nil
      gx = @template.gauss_gradient_x @sigma, 0.01
      gy = @template.gauss_gradient_y @sigma, 0.01
      gx = gx[ @fringe...(@template.shape[0]-@fringe),
               @fringe...(@template.shape[1]-@fringe) ]
      gy = gy[ @fringe...(@template.shape[0]-@fringe),
               @fringe...(@template.shape[1]-@fringe) ]
      tpl = @template[ @fringe...(@template.shape[0]-@fringe),
                       @fringe...(@template.shape[1]-@fringe) ]
      w, h = *tpl.shape
      idx = MultiArray.int( w, h ).indgen!
      x, y = idx % w, idx / w
      c = derivative( x, y ) * Vector[ gx, gy ]
      hinv = ( ( c * c.covector ).collect { |e| e.sum } ).inv
      diff = nil
      pose = @model.pose * @delta.inv
      @iterations.times do
        warp = MultiArray.sfloat( w, h, 2 )
        warp[ 0...w, 0...h, 0 ] =
          pose[ 0, 0 ] * x + pose[ 0, 1 ] * y + pose[ 0, 2 ]
        warp[ 0...w, 0...h, 1 ] =
          pose[ 1, 0 ] * x + pose[ 1, 1 ] * y + pose[ 1, 2 ]
        diff = image.warp_clipped_interpolate( warp ) - tpl.to_dfloat
        s = c.collect { |e| ( e * diff ).sum }
        pose = compose pose, hinv * s
      end
      @model.pose = pose * @delta
      @deviation = Math.sqrt( ( diff * diff ).sum / diff.size )
    end
  end

end
