require 'visionbase'
require 'ui_plugin_ncc'

class TemplateWidget < ImageDisplay

  signals 'capture(int,int)'

  def initialize( parent = nil )
    super parent
    @selection, @selecting, @image = false, false, nil
    @x0, @y0, @w, @h = 0, 0, 0, 0
  end

  def write( image )
    @image = image
    super image
  end

  def paintEvent( e )
    super e
    if @selection or @selecting
      p = Qt::Painter.new self
      p.setRenderHint Qt::Painter::Antialiasing, false
      pen = Qt::Pen.new Qt::Color.new( 255, 0, 0 )
      pen.style = Qt::DotLine if @selecting
      p.pen = pen
      p.drawRect @x0, @y0, @w - 1, @h - 1
      p.end
    end
  end

  def mousePressEvent( e )
    if @image != nil
      case e.button
      when Qt::LeftButton
        clear
        @x0, @y0, @w, @h = e.x, e.y, 1, 1
        @selection, @selecting = false, true
        repaint @x0, @y0, @w, @h
      end
    end
  end

  def mouseMoveEvent( e )
    if @selecting
      x = [ @x0, @x0 + @w - 1 ]
      y = [ @y0, @y0 + @h - 1 ]
      @w = [ [ @image.shape[0] - @x0, e.x + 1 - @x0 ].min, 1 ].max
      @h = [ [ @image.shape[1] - @y0, e.y + 1 - @y0 ].min, 1 ].max
      x.push @x0 + @w - 1
      y.push @y0 + @h - 1
      repaint x.min, y.min, x.max + 1 - x.min, y.max + 1 - y.min
    end
  end

  def mouseReleaseEvent( e )
    if @selecting
      case e.button
      when Qt::LeftButton
        @selection, @selecting = true, false
        repaint @x0, @y0, @w, @h
        emit capture( @x0, @y0 )
      end
    end
  end

  def selection( x0, y0, w, h )
    @x0, @y0, @w, @h =  x0, y0, w, h
    @selection = true
    repaint
  end

  def clear
    if @selection or @selecting
      x = [ @x0, @x0 + @w - 1 ]
      y = [ @y0, @y0 + @h - 1 ]
      @selection, @selecting = false, false
      repaint @x0, @y0, @w, @h
    end
  end

  def template
    if @selection
      @image[ @x0...(@x0+@w), @y0...(@y0+@h) ]
    else
      nil
    end
  end

end

class FormNcc < AlgorithmForm

  slots 'noise(double)'
  slots 'threshold(double)'
  slots 'capture(int,int)'

  def initialize( parent = nil )
    super parent
    @ui = Ui::FormNcc.new
    @ui.setupUi self
    @template_widget = TemplateWidget.new
    @ui.scroll_area.widget = @template_widget
    connect @ui.noise_spin, SIGNAL('valueChanged(double)'),
      self, SLOT('noise(double)')
    connect @ui.threshold_spin, SIGNAL('valueChanged(double)'),
      self, SLOT('threshold(double)')
    connect @template_widget, SIGNAL('capture(int,int)'),
      self, SLOT('capture(int,int)')
  end

  def image=( image )
    @template_widget.write image
  end

  def algorithm=( algorithm )
    super algorithm
    @ui.noise_spin.value = algorithm.noise
    @ui.threshold_spin.value = algorithm.threshold
    if algorithm.template != nil
      pose = @algorithm.model.pose * @algorithm.delta.inv
      @template_widget.selection pose[ 0, 2 ], pose[ 1, 2 ],
                                 *algorithm.template.shape
    else
      @template_widget.clear
    end
  end

  def noise( value )
    @algorithm.noise = value
  end

  def threshold( value )
    @algorithm.threshold = value
  end

  def capture( x, y )
    @algorithm.template = @template_widget.template
    @algorithm.model.pose = Pose.new( x, y ) * @algorithm.delta
    @algorithm.max_val = 1.0
  end

end

class AlgorithmNcc < Recognition

  class << self

    def description
      'Normalised Cross-Correlation'
    end

    def config_widget( parent = nil )
      FormNcc.new parent
    end

  end

  attr_accessor :noise, :threshold, :template, :max_val

  def initialize( model = nil )
    super model
    @noise, @threshold, @template = 1.0, 0.9, nil
    @max_val = 0
  end

  Struct.new 'AlgorithmNcc', :base, :noise, :threshold, :typecode, :export,
    :shape, :max_val

  def marshal_load( str )
    super str.base
    @noise, @threshold, @max_val = str.noise, str.threshold, str.max_val
    @template = MultiArray.import str.typecode, str.export, *str.shape
    self
  end

  def marshal_dump
    Struct::AlgorithmNcc.new super, @noise, @threshold, @template.typecode,
      @template.export, @template.shape, @max_val
  end

  def overlay( widget )
    if @template != nil
      p = Qt::Painter.new widget
      p.setRenderHint Qt::Painter::Antialiasing, false
      pen = Qt::Pen.new Qt::Color.new( 255, 0, 0 )
      pen.style = Qt::DotLine unless success?
      p.pen = pen
      p.setWorldTransform( ( @model.pose * @delta.inv ).to_qtransform )
      w, h = *@template.shape
      p.drawRect 0, 0, w - 1, h - 1
      p.end
    end
  end

  def success?
    @max_val >= @threshold
  end

  def update( image )
    if @template != nil
      ncc = image.to_dfloat.ncc @template.to_dfloat, @noise
      @max_val = ncc.max
      idx = MultiArray.int( *ncc.shape ).indgen!.mask( ncc >= @max_val )
      unless idx.empty? # deal with NaN
        x0 = idx[0] % ncc.shape[0]
        y0 = idx[0] / ncc.shape[0]
        x0 = 0 if x0 < 0
        y0 = 0 if y0 < 0
        w, h = *@template.shape
        x0 = image.shape[0] - w if x0 + w > image.shape[0]
        y0 = image.shape[1] - h if y0 + h > image.shape[1]
        @model.pose = Pose.new( x0, y0 ) * @delta
      end
    end
  end

end

