require 'hornetseye'
require 'matrix'
require 'matrix_fix'
require 'Qt4'

module Magick

  class Image

    def to_pixmap
      str = to_blob { self.format = "PPM"; self.depth = 8 }
      pix = Qt::Pixmap.new
      pix.loadFromData Qt::ByteArray.fromRawData( str, str.size )
      pix
    end

    def shape
      [ columns, rows ]
    end

  end

end

module Hornetseye

  class Sequence_

    def integral
      y = 0
      collect { |x| retval = y; y += x; retval }
    end

    def inverse
      hist_weighted size, self.class.indgen
    end

    def gauss_gradient_x( sigma, max_error = 1.0 / 256.0 )
      gauss_gradient sigma, 0, max_error
    end

    def gauss_gradient_y( sigma, max_error = 1.0 / 256.0 )
      gauss_gradient sigma, 1, max_error
    end

    def to_pixmap
      to_ubytergb.to_magick.to_pixmap
    end

    def avg
      to_dfloat.sum / size
    end

    def sqr
      self * self
    end

    def warp_clipped_interpolate( field )
      shape = field.shape[ 0...field.shape.size-1 ]
      field00 = field.floor.to_int
      field01 = field00.dup
      field01[ 0...shape[0], 0...shape[1], 0 ] += 1
      field10 = field00.dup
      field10[ 0...shape[0], 0...shape[1], 1 ] += 1
      field11 = field00.dup
      field11 = field11 + 1
      fieldfrac = field - field00
      frac00 = fieldfrac[ 0...shape[0], 0...shape[1], 0 ]
      frac01 = 1 - fieldfrac[ 0...shape[0], 0...shape[1], 0 ]
      frac10 = fieldfrac[ 0...shape[0], 0...shape[1], 1 ]
      frac11 = 1 - fieldfrac[ 0...shape[0], 0...shape[1], 1 ]
      return warp_clipped( field00 ) * frac01 * frac11 +
        warp_clipped( field01 ) * frac00 * frac11 +
        warp_clipped( field10 ) * frac01 * frac10 +
        warp_clipped( field11 ) * frac00 * frac10
    end

    def corr( other )
      ( rfft * other.rfft.conj ).irfft
    end

    def zcorr( other )
      zother = MultiArray.dfloat( *shape ).fill!
      zother[ 0...other.shape[0], 0...other.shape[1] ] = other
      corr( zother )
    end

    def median( *shape )
      filter = MultiArray.dfloat( *shape ).fill!( 1 )
      zcorr( filter )
    end

    def ncc( other, noise )
      zcorr( other - other.avg ) /
        Math.sqrt( ( sqr.median( *other.shape ) -
                     median( *other.shape ).sqr / other.size ) *
                   ( other - other.avg ).sqr.sum + noise )

    end

    def edges( sigma = 1.0, error = 1.0 / 16.0 )
      ( ( gauss_gradient_x( sigma, error ).sqr +
          gauss_gradient_y( sigma, error ).sqr ) >= 9.0 ** 2 ).to_ubyte
    end

  end

end

if false
class VideoDisplay < Hornetseye::XvWidget

  attr_accessor :draw_x11

  def initialize( parent = nil, grab_port_now = true )
    super parent, grab_port_now
    @image, @pixmap, @draw_x11 = nil, nil, true
  end

  def write( image )
    @image, @pixmap = image, nil
    resize *image.shape
    super image
  end
  
  def paintEvent( e )
    if @draw_x11 and @image != nil
      @pixmap = @image.to_pixmap if @pixmap == nil
      p = Qt::Painter.new self
      p.drawPixmap 0, 0, @image.to_pixmap
      p.end
    else
      super e
    end
  end

  def clear
    @image = nil
    super
  end

end
end

class ImageDisplay < Qt::Label

  def write( image )
    setPixmap image.to_pixmap
    resize *image.shape
    repaint
  end
  
end

class VisionDisplay < ImageDisplay

  attr_accessor :models

  def initialize( parent = nil )
    super parent #, false
    @models = []
  end

  def paintEvent( e )
    super e
    @models.each { |model| model.overlay self }
  end

end

class CoordDisplay < ImageDisplay

  signals 'selected()'

  attr_reader :pose

  def initialize( parent = nil )
    super parent
    @selected, @pose = false, Pose.new
  end

  def paintEvent( e )
    super e
    if @selected
      p = Qt::Painter.new self
      p.setRenderHint Qt::Painter::Antialiasing, false
      p.pen = Qt::Pen.new Qt::Color.new( 0, 0, 255 )
      p.setWorldTransform @pose.to_qtransform
      p.drawLine -5, 0, 50, 0
      p.drawLine 50, 0, 45, -5
      p.drawLine 50, 0, 45, 5
      p.drawLine 0, -5, 0, 50
      p.drawLine -5, 45, 0, 50
      p.drawLine 0, 50, 5, 45
      p.end
    end
  end

  def mousePressEvent( e )
    case e.button
    when Qt::LeftButton
      select Pose.new( e.x, e.y )
    end
  end

  def wheelEvent( e )
    if @selected
      angle = Math.atan2( @pose[ 1, 0 ], @pose[ 0, 0 ] )
      delta = e.delta / 8.0 * 2 * Math::PI / 360.0
      @pose = Pose.new( @pose[ 0, 2 ], @pose[ 1, 2 ], angle + delta )
      repaint @pose[ 0, 2 ] - 50, @pose[ 1, 2 ] - 50, 100, 100
      emit selected
    end
  end

  def select( v )
    @selected = false
    repaint @pose[ 0, 2 ] - 50, @pose[ 1, 2 ] - 50, 100, 100
    @selected, @pose = true, v
    repaint @pose[ 0, 2 ] - 50, @pose[ 1, 2 ] - 50, 100, 100
    emit selected
  end

end

class Pose

  attr_accessor :matrix

  def initialize( t1 = 0.0, t2 = 0.0, w = 0.0 )
    cw, sw = Math.cos( w ), Math.sin( w )
    @matrix = Matrix[ [ cw, -sw, t1 ], [ sw, cw, t2 ], [ 0.0, 0.0, 1.0 ] ]
  end

  def inspect
    "Pose(#{@matrix[0,2]},#{@matrix[1,2]},#{Math.atan2(@matrix[1,0],@matrix[0,0])})"
  end

  def []( row, column )
    @matrix[ row, column ]
  end

  def []=( row, column, value )
    arr = @matrix.to_a
    arr[ row ][ column ] = value
    @matrix = Matrix[ *arr ]
  end

  def *( other )
    retval = Pose.new
    retval.matrix = @matrix * other.matrix
    retval
  end

  def inv
    retval = Pose.new
    retval.matrix = @matrix.inv
    retval
  end

  def to_qtransform
    Qt::Transform.new( @matrix[ 0, 0 ], @matrix[ 1, 0 ],
                       @matrix[ 0, 1 ], @matrix[ 1, 1 ],
                       @matrix[ 0, 2 ], @matrix[ 1, 2 ] )
  end

end

class Model

  attr_reader :recognition_algorithm, :tracking_algorithm
  attr_accessor :name, :pose, :tracking

  def initialize( name, recognition_algorithm, tracking_algorithm = nil )
    @name, @recognition_algorithm, @tracking_algorithm, @tracking =
      name, recognition_algorithm, tracking_algorithm, false
    @pose = Pose.new
    recognition_algorithm.model = self if recognition_algorithm != nil
    tracking_algorithm.model = self if tracking_algorithm != nil
  end

  def recognition_algorithm=( algorithm )
    algorithm.model = self
    @recognition_algorithm = algorithm
  end

  def tracking_algorithm=( algorithm )
    algorithm.model = self if algorithm != nil
    @tracking_algorithm = algorithm
  end

  def overlay( widget )
    if @tracking and @tracking_algorithm.success?
      @tracking_algorithm.overlay widget
    else
      @recognition_algorithm.overlay widget
    end
    if success?
      p = Qt::Painter.new widget
      p.setRenderHint Qt::Painter::Antialiasing, false
      p.pen = Qt::Pen.new Qt::Color.new( 0, 0, 255 )
      p.setWorldTransform @pose.to_qtransform
      p.drawLine -5, 0, 50, 0
      p.drawLine 50, 0, 45, -5
      p.drawLine 50, 0, 45, 5
      p.drawLine 0, -5, 0, 50
      p.drawLine -5, 45, 0, 50
      p.drawLine 0, 50, 5, 45
      p.setWorldTransform Qt::Transform.new( 1.0, 0.0, 0.0, 1.0,
                                             @pose[ 0, 2 ], @pose[ 1, 2 ] )
      rect = p.boundingRect( 50, 30, 80, 20, 0, @name )
      p.drawLine 0, 0, rect.x, rect.y + rect.height
      p.drawLine rect.x, rect.y + rect.height,
        rect.x + rect.width - 1, rect.y + rect.height
      p.fillRect rect, Qt::Brush.new( Qt::white )
      p.drawText 50, 30, 80, 20, 0, @name 
      p.end
    end
  end

  def success?
    if @tracking and @tracking_algorithm.success?
      true
    else
      @recognition_algorithm.success?
    end
  end

  def log( csv, timestamp )
    if success?
      csv << [ '%10.4f' % timestamp,
               '%10.4f' % @pose[ 0, 2 ],
               '%10.4f' % @pose[ 1, 2 ],
               '%10.4f' % ( Math.atan2( @pose[ 1, 0 ],
                                        @pose[ 0, 0 ] ) * 180.0 / Math::PI ) ] 
    end
  end

end

class AlgorithmForm < Qt::Widget

  attr_accessor :algorithm

  def initialize( parent = nil )
    super parent
    @algorithm = nil
  end

  def image=( image )
  end

end

class Algorithm

  class << self

    def config_widget( parent = nil )
      AlgorithmForm.new parent
    end

  end

  attr_accessor :delta, :model

  def initialize( model = nil )
    @model = model
    @delta = Pose.new
  end

  Struct.new 'Algorithm', :delta, :model

  def marshal_load( str )
    @delta = str.delta
    @model = str.model
    self
  end

  def marshal_dump
    Struct::Algorithm.new @delta, @model
  end

  def overlay( widget )
  end

  def success?
    true
  end

  def update( image )
  end

end

class Recognition < Algorithm

  class << self

    def description
      'recognition template'
    end

  end

end

class Tracking < Algorithm

  class << self

    def description
      'tracking template'
    end

  end

end
