require 'Qt4'
require 'qrc_visiongui'
require 'ui_visionwindow'
require 'hornetseye'
require 'modeldialog'
require 'csv'

Dir.glob( 'plugin_*.rb' ).each { |file| require file }

class VisionWindow < Qt::MainWindow

  attr_reader :image

  STATUSDELAY = 5000

  slots 'open_file()'
  slots 'open_camera()'
  slots 'logging()'
  slots 'play(bool)'
  slots 'gui_stop()'
  slots 'config()'
  slots 'conf_save()'
  slots 'conf_save_as()'
  slots 'conf_load()'

  def initialize( parent = nil )
    super parent

    # Setup GUI.
    @ui = Ui::VisionWindow.new
    @ui.setupUi self

    # Setup video, timer, and list of objects to recognise/track, ...
    @image = nil
    @video = nil
    @timer = 0
    @models = []
    @video_file = nil
    @conf_file = nil
    @log_file = nil
    @csv_writer = {}

    # Insert video widget.
    @video_display = VisionDisplay.new self
    @video_display.models = @models
    @video_display.resize 320, 240
    @ui.scroll_area.widget = @video_display

    # Connect controls.
    connect @ui.action_openfile, SIGNAL('triggered()'),
      self, SLOT('open_file()')
    connect @ui.action_opencamera, SIGNAL('triggered()'),
      self, SLOT('open_camera()')
    connect @ui.action_csv, SIGNAL('triggered()'),
      self, SLOT('logging()')
    connect @ui.action_play, SIGNAL('toggled(bool)'), self, SLOT('play(bool)')
    connect @ui.action_stop, SIGNAL('triggered()'), self, SLOT('gui_stop()')
    connect @ui.action_config, SIGNAL('triggered()'), self, SLOT('config()')
    connect @ui.action_load_conf, SIGNAL('triggered()'),
      self, SLOT('conf_load()')
    connect @ui.action_save_conf, SIGNAL('triggered()'),
      self, SLOT('conf_save()')
    connect @ui.action_save_conf_as, SIGNAL('triggered()'),
      self, SLOT('conf_save_as()')
    connect @ui.action_quit, SIGNAL('triggered()'), self, SLOT('close()')

    # Load plugins.
    @recognition_algorithms = []
    @tracking_algorithms = []
    Dir.glob( 'plugin_*.rb' ).sort.each do |file|
      name = file.gsub /plugin_(.*)\.rb/, '\1'
      plugin = Object.const_get "Algorithm#{name.capitalize}"
      if plugin < Recognition
        @recognition_algorithms.push plugin
      elsif plugin < Tracking
        @tracking_algorithms.push plugin
      else
        raise "#{plugin.inspect} must inherit either Recognition or Tracking"
      end
    end
    if @recognition_algorithms.empty?
      raise 'At least one plugin (plugin_*.rb) providing a recognition ' +
        'algorithm is required'
    end

    # Allow opening video with command line argument.
    play_video ARGV[0] if ARGV.size == 1

    # Signal ready.
    statusBar.showMessage 'Ready', STATUSDELAY
  end

  def closeEvent( e )
    @ui.action_play.checked = false
    decision =
      Qt::MessageBox::question self, 'Close Vision GUI',
                               'Exit the application?',
                               Qt::MessageBox::Ok,
                               Qt::MessageBox::Cancel
    case decision
    when Qt::MessageBox::Ok
      stop
      e.accept
    when Qt::MessageBox::Cancel
      e.ignore
    end
  end

  def open_file
    begin
      @ui.action_play.checked = false
      file = Qt::FileDialog.getOpenFileName self, 'Load Video',
                                            @video_file,
                                            'AVI Files (*.avi);;' +
                                            'MPEG Files (*.mpg *.mp4);;' +
                                            'All Files (*)'
      if file
        statusBar.showMessage "Opening #{file}", STATUSDELAY
        play_video file
      else
        statusBar.showMessage 'Aborted loading of video', STATUSDELAY
      end
    rescue Exception => e
      statusBar.showMessage e.to_s, STATUSDELAY
      Qt::MessageBox::critical self, 'Error Opening Video', e.to_s
    end
  end

  def open_camera
    begin
      @ui.action_play.checked = false
      dialog = CameraDialog.new self
      if dialog.exec == Qt::Dialog::Rejected
        statusBar.showMessage 'Aborted camera dialog', STATUSDELAY
      else
        statusBar.showMessage 'Accessing firewire digital camera', STATUSDELAY
        access_camera dialog.device, dialog.node, dialog.channel,
          dialog.format, dialog.mode, dialog.framerate
      end
    rescue Exception => e
      statusBar.showMessage e.to_s, STATUSDELAY
      Qt::MessageBox::critical self, 'Error Accessing Camera', e.to_s
    end
  end

  def logging
    begin
      dialog = ModelDialog.new self
      dialog.model_names = @models.collect { |m| m.name }
      if dialog.exec == Qt::Dialog::Accepted
        model_name = dialog.model
        file = Qt::FileDialog.getSaveFileName self, 'Save Configuration',
                                              @log_file,
                                              'CSV Files (*.csv);;' +
                                              'All files (*)'
        if file
          statusBar.showMessage "Creating CSV file #{file} for model #{model_name}", STATUSDELAY
          @log_file = file
          f = File.new file, 'w'
          w = CSV::Writer.create f, "\t"
          w << [ 'Time/s', 'x/pixel', 'y/pixel', 'angle' ]
          @csv_writer[ model_name ] = [ f, w ]
        else
          statusBar.showMessage "Aborted creation of CSV file for model #{model_name}", STATUSDELAY
        end
      else
        statusBar.showMessage 'Aborted creation of new CSV file', STATUSDELAY
      end
    rescue Exception => e
      statusBar.showMessage e.to_s, STATUSDELAY
      Qt::MessageBox::critical self, 'Error Creating CSV File', e.to_s
    end
  end

  def play_video( file )
    # @video = XineInput.new file
    @video = MPlayerInput.new file
    @video.instance_eval do
      def pos
        @pos
      end
      @pos = -1
      def read_ubyte
        @pos = @pos + 1
        read.to_ubyte
      end
    end
    @video_file = file
    @conf_file = nil
    @image = @video.read_ubyte
    @models = []
    @video_display.models = @models
    # @video_display.grabPort
    @video_display.write @image
    @video_display.repaint
    @ui.action_openfile.enabled = false
    @ui.action_opencamera.enabled = false
    @ui.action_csv.enabled = false
    @ui.action_play.enabled = true
    @ui.action_stop.enabled = true
    @ui.action_config.enabled = true
    @ui.action_load_conf.enabled = true
    @ui.action_save_conf.enabled = false
    @ui.action_save_conf_as.enabled = true
  end

  def access_camera( device, node, channel, format, mode, framerate )
    @video = DC1394Input.new( device, node, channel, format, mode, -1,
                              framerate, DC1394Input::SPEED_400 )
    @t0 = Time.new.to_f
    @conf_file = nil
    @image = @video.read_ubyte
    @models = []
    @video_display.models = @models
    #@video_display.grabPort
    @video_display.write @image
    @video_display.repaint
    @ui.action_openfile.enabled = false
    @ui.action_opencamera.enabled = false
    @ui.action_csv.enabled = false
    @ui.action_play.enabled = true
    @ui.action_stop.enabled= true
    @ui.action_config.enabled = true
    @ui.action_load_conf.enabled = true
    @ui.action_save_conf.enabled = false
    @ui.action_save_conf_as.enabled = true
    @ui.action_play.checked = true
  end

  def play( on )
    if on
      @ui.action_config.enabled = false
      @ui.action_load_conf.enabled = false
      @ui.action_save_conf.enabled = false
      @ui.action_save_conf_as.enabled = false
      # @timer = startTimer @video.frame_duration / 90 if @timer == 0
      @timer = startTimer 0 if @timer == 0
      # @video_display.draw_x11 = false
    else
      killTimer @timer if @timer != 0
      @timer = 0
      @ui.action_config.enabled = true
      @ui.action_load_conf.enabled = true
      @ui.action_save_conf.enabled = @conf_file != nil
      @ui.action_save_conf_as.enabled = true
      # @video_display.draw_x11 = true
      @video_display.repaint
    end
  end

  def gui_stop
    @ui.action_play.checked = false
    decision = 
      Qt::MessageBox::question self, 'Stop Video',
                               'Reset video input and configuration?',
                               Qt::MessageBox::Ok,
                               Qt::MessageBox::Cancel
    case decision
    when Qt::MessageBox::Ok
      stop
      statusBar.showMessage 'Video input and configuration was reset', STATUSDELAY
    when Qt::MessageBox::Cancel
      statusBar.showMessage 'Reset of video input and configuration was aborted', STATUSDELAY
    end
  end

  def stop
    @ui.action_play.checked = false
    @video.close if @video != nil
    @conf_file = nil
    @models = []
    @video_display.models = @models
    @video_display.clear
    #@video_display.releasePort
    @image = nil
    @video = nil
    @csv_writer.each { |n,csv| csv[0].close }
    @csv_writer = {}
    @ui.action_openfile.enabled = true
    @ui.action_opencamera.enabled = true
    @ui.action_csv.enabled = false
    @ui.action_play.enabled = false
    @ui.action_stop.enabled = false
    @ui.action_config.enabled = false
    @ui.action_load_conf.enabled = false
    @ui.action_save_conf.enabled = false
    @ui.action_save_conf_as.enabled = false
  end

  def config
    # Create copy of settings.
    old_data = Marshal.dump @models

    # Create configuration dialog.
    dialog = VisionDialog.new self,
                              @recognition_algorithms, @tracking_algorithms
    dialog.image = @image
    dialog.models = @models
    if dialog.exec == Qt::Dialog::Rejected
      statusBar.showMessage 'Aborted configuration dialog', STATUSDELAY
      @models = Marshal.load old_data
    else
      statusBar.showMessage 'Applying new configuration', STATUSDELAY
      # Force recognition in next step.
      @models.each { |model| model.tracking = false }
    end
    @ui.action_csv.enabled = !@models.empty?
    @video_display.models = @models
    @video_display.repaint
  end

  def conf_load
    begin
      file = Qt::FileDialog.getOpenFileName self, 'Load Configuration',
                                            @conf_file,
                                            'Config Files (*.dat);;' +
                                            'All files (*)'
      if file
        statusBar.showMessage "Loading #{file}", STATUSDELAY
        load_conf file
      else
        statusBar.showMessage 'Aborted loading of configuration', STATUSDELAY
      end
    rescue Exception => e
      statusBar.showMessage e.to_s, STATUSDELAY
      Qt::MessageBox::critical self, 'Error Loading Configuration', e.to_s
    end
  end

  def conf_save
    begin
      statusBar.showMessage "Saving #{@conf_file}", STATUSDELAY
      save_conf @conf_file
    rescue Exception => e
      statusBar.showMessage e.to_s, STATUSDELAY
      Qt::MessageBox::critical self, 'Error Saving Configuration', e.to_s
    end
  end

  def conf_save_as
    begin
      file = Qt::FileDialog.getSaveFileName self, 'Save Configuration',
                                            @conf_file,
                                            'Config Files (*.dat);;' +
                                            'All files (*)'
      if file
        statusBar.showMessage "Saving #{file}", STATUSDELAY
        save_conf file
      else
        statusBar.showMessage 'Aborted saving of configuration', STATUSDELAY
      end
    rescue Exception => e
      statusBar.showMessage e.to_s, STATUSDELAY
      Qt::MessageBox::critical self, 'Error Saving Configuration', e.to_s
    end
  end

  def load_conf( file )
    f = File.new file, 'rb'
    @models = Marshal.load f.read
    # Force recognition in next step.
    @models.each { |model| model.tracking = false }
    @video_display.models = @models
    @video_display.repaint
    @conf_file = file
    @ui.action_csv.enabled = !@models.empty?
    @ui.action_save_conf.enabled = true
  end

  def save_conf( file )
    f = File.new file, 'wb'
    f.write Marshal.dump( @models )
    f.close
    @conf_file = file
    @ui.action_save_conf.enabled = true
  end

  def timerEvent( event )
    case event.timerId
    when @timer
      begin
        @image = @video.read_ubyte # need dup in some cases !!!
      rescue Exception => e
        stop
      end
      if @image != nil
        @models.each do |model|
          if model.tracking_algorithm != nil
            if model.tracking
              model.tracking_algorithm.update @image
              model.tracking = false unless model.tracking_algorithm.success?
            end
            unless model.tracking
              model.recognition_algorithm.update @image
              model.tracking = true if model.recognition_algorithm.success?
            end
          else
            model.recognition_algorithm.update @image
          end
          csv = @csv_writer[ model.name ]
          if csv != nil
            if @video.respond_to? :pos
              model.log csv[1], @video.pos * 0.001
            else
              model.log csv[1],
                        ( ( Time.new.to_f - @t0 ) * 1000 ).round * 0.001
            end
          end
        end
        @video_display.write @image
        @video_display.repaint
      end
    end
  end

end
